import allure
import pytest
from frostfs_testlib import reporter
from frostfs_testlib.cli.frostfs_cli.cli import FrostfsCli
from frostfs_testlib.resources.error_patterns import OBJECT_ACCESS_DENIED
from frostfs_testlib.steps.cli.object import get_object_from_random_node, head_object, put_object_to_random_node
from frostfs_testlib.storage.dataclasses import ape
from frostfs_testlib.storage.dataclasses.wallet import WalletInfo
from frostfs_testlib.testing.cluster_test_base import ClusterTestBase
from frostfs_testlib.testing.test_control import expect_not_raises
from frostfs_testlib.utils.file_utils import TestFile

from ....helpers.bearer_token import create_bearer_token
from ....helpers.container_access import (
    ALL_OBJECT_OPERATIONS,
    FULL_ACCESS,
    assert_access_to_container,
    assert_full_access_to_container,
    assert_no_access_to_container,
)
from ....helpers.container_spec import ContainerSpec
from ....helpers.object_access import OBJECT_ACCESS_DENIED


@pytest.mark.nightly
@pytest.mark.ape
class TestApeFilters(ClusterTestBase):
    #  SPEC: https://github.com/nspcc-dev/neofs-spec/blob/master/01-arch/07-acl.md
    HEADER = {"check_key": "check_value"}
    OTHER_HEADER = {"check_key": "other_value"}
    ATTRIBUTES = {
        "key_one": "check_value",
        "x_key": "xvalue",
        "check_key": "check_value",
    }
    OTHER_ATTRIBUTES = {
        "key_one": "check_value",
        "x_key": "other_value",
        "check_key": "other_value",
    }

    OBJECT_COUNT = 5
    RESOURCE_OPERATIONS = [
        ape.ObjectOperations.GET,
        ape.ObjectOperations.HEAD,
        ape.ObjectOperations.PUT,
    ]

    @pytest.fixture
    def objects_with_attributes(self, default_wallet: WalletInfo, file_path: TestFile, container: str):
        return [
            put_object_to_random_node(
                default_wallet, file_path, container, self.shell, self.cluster, attributes={**self.ATTRIBUTES, "key": val}
            )
            for val in range(self.OBJECT_COUNT)
        ]

    @pytest.fixture
    def objects_with_other_attributes(self, default_wallet: WalletInfo, file_path: TestFile, container: str):
        return [
            put_object_to_random_node(
                default_wallet, file_path, container, self.shell, self.cluster, attributes={**self.OTHER_ATTRIBUTES, "key": val}
            )
            for val in range(self.OBJECT_COUNT)
        ]

    @pytest.fixture
    def objects_without_attributes(self, default_wallet: WalletInfo, file_path: TestFile, container: str):
        return [put_object_to_random_node(default_wallet, file_path, container, self.shell, self.cluster) for _ in range(self.OBJECT_COUNT)]

    @pytest.mark.sanity
    @allure.title("Operations with request filter (match_type={match_type}, obj_size={object_size})")
    @pytest.mark.parametrize("match_type", [ape.MatchType.EQUAL, ape.MatchType.NOT_EQUAL])
    @pytest.mark.skip("https://git.frostfs.info/TrueCloudLab/frostfs-node/issues/1243")
    def test_ape_filters_request(
        self,
        frostfs_cli: FrostfsCli,
        temp_directory: str,
        other_wallet: WalletInfo,
        container: str,
        objects_with_attributes: list[str],
        objects_with_other_attributes: list[str],
        objects_without_attributes: list[str],
        match_type: ape.MatchType,
        file_path: TestFile,
        rpc_endpoint: str,
    ):
        with reporter.step("Deny all operations for others via APE with request condition"):
            request_condition = ape.Condition('"frostfs:xheader/check_key"', '"check_value"', ape.ConditionType.REQUEST, match_type)
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            deny_rule = ape.Rule(ape.Verb.DENY, ALL_OBJECT_OPERATIONS, [request_condition, role_condition])

            frostfs_cli.ape_manager.add(
                rpc_endpoint, deny_rule.chain_id, target_name=container, target_type="container", rule=deny_rule.as_string()
            )

        with reporter.step("Wait for one block"):
            self.wait_for_blocks()

        with reporter.step("Create bearer token with everything allowed for others role"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            rule = ape.Rule(ape.Verb.ALLOW, ALL_OBJECT_OPERATIONS, role_condition)
            bearer = create_bearer_token(frostfs_cli, temp_directory, container, rule, rpc_endpoint)

        # Filter denies requests where "check_key {match_type} ATTRIBUTE", so when match_type
        # is STRING_EQUAL, then requests with "check_key=OTHER_ATTRIBUTE" will be allowed while
        # requests with "check_key=ATTRIBUTE" will be denied, and vice versa
        allow_headers = self.OTHER_HEADER if match_type == ape.MatchType.EQUAL else self.HEADER
        deny_headers = self.HEADER if match_type == ape.MatchType.EQUAL else self.OTHER_HEADER

        # We test on 3 groups of objects with various headers,
        # but APE rule should ignore object headers and only work based on request headers
        for oids in [objects_with_attributes, objects_with_other_attributes, objects_without_attributes]:
            with reporter.step("Check others has full access when sending request without headers"):
                assert_full_access_to_container(other_wallet, container, oids.pop(), file_path, self.shell, self.cluster)

            with reporter.step("Check others has full access when sending request with allowed headers"):
                assert_full_access_to_container(
                    other_wallet, container, oids.pop(), file_path, self.shell, self.cluster, xhdr=allow_headers
                )

            with reporter.step("Check others has no access when sending request with denied headers"):
                assert_no_access_to_container(other_wallet, container, oids.pop(), file_path, self.shell, self.cluster, xhdr=deny_headers)

            with reporter.step("Check others has full access when sending request with denied headers and using bearer token"):
                assert_full_access_to_container(
                    other_wallet, container, oids.pop(), file_path, self.shell, self.cluster, bearer, deny_headers
                )

    @allure.title("Operations with deny user headers filter (match_type={match_type}, obj_size={object_size})")
    @pytest.mark.parametrize("match_type", [ape.MatchType.EQUAL, ape.MatchType.NOT_EQUAL])
    @pytest.mark.skip("https://git.frostfs.info/TrueCloudLab/frostfs-node/issues/1300")
    def test_ape_deny_filters_object(
        self,
        frostfs_cli: FrostfsCli,
        temp_directory: str,
        other_wallet: WalletInfo,
        container: str,
        objects_with_attributes: list[str],
        objects_with_other_attributes: list[str],
        objects_without_attributes: list[str],
        match_type: ape.MatchType,
        rpc_endpoint: str,
        file_path: TestFile,
    ):
        allow_objects = objects_with_other_attributes if match_type == ape.MatchType.EQUAL else objects_with_attributes
        deny_objects = objects_with_attributes if match_type == ape.MatchType.EQUAL else objects_with_other_attributes

        # When there is no attribute on the object, it's the same as "", and "" is not equal to "<some_value>"
        # So it's the same as deny_objects
        no_attributes_access = {
            ape.MatchType.EQUAL: FULL_ACCESS,
            ape.MatchType.NOT_EQUAL: {
                ape.ObjectOperations.PUT: False,
                ape.ObjectOperations.GET: False,
                ape.ObjectOperations.HEAD: False,
                ape.ObjectOperations.GET_RANGE: True,
                ape.ObjectOperations.GET_RANGE_HASH: True,
                ape.ObjectOperations.SEARCH: True,
                ape.ObjectOperations.DELETE: False,  # Denied by restricted PUT
            },
        }
        allowed_access = {
            ape.MatchType.EQUAL: FULL_ACCESS,
            ape.MatchType.NOT_EQUAL: {
                ape.ObjectOperations.PUT: False,  # because currently we are put without attributes
                ape.ObjectOperations.GET: True,
                ape.ObjectOperations.HEAD: True,
                ape.ObjectOperations.GET_RANGE: True,
                ape.ObjectOperations.GET_RANGE_HASH: True,
                ape.ObjectOperations.SEARCH: True,
                ape.ObjectOperations.DELETE: False,  # Because delete needs to put a tombstone without attributes
            },
        }

        with reporter.step("Deny operations for others via APE with resource condition"):
            resource_condition = ape.Condition('"check_key"', '"check_value"', ape.ConditionType.RESOURCE, match_type)
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            deny_rule = ape.Rule(ape.Verb.DENY, self.RESOURCE_OPERATIONS, [resource_condition, role_condition])

            frostfs_cli.ape_manager.add(
                rpc_endpoint, deny_rule.chain_id, target_name=container, target_type="container", rule=deny_rule.as_string()
            )

        with reporter.step("Wait for one block"):
            self.wait_for_blocks()

        with reporter.step("Create bearer token with everything allowed for others role"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            rule = ape.Rule(ape.Verb.ALLOW, ALL_OBJECT_OPERATIONS, role_condition)
            bearer = create_bearer_token(frostfs_cli, temp_directory, container, rule, rpc_endpoint)

        with reporter.step("Create bearer token with allowed put for others role"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            rule = ape.Rule(ape.Verb.ALLOW, ape.ObjectOperations.PUT, role_condition)
            bearer_put = create_bearer_token(frostfs_cli, temp_directory, container, rule, rpc_endpoint)

        # We will attempt requests with various headers,
        # but APE rule should ignore request headers and validate only object headers
        for xhdr in (self.HEADER, self.OTHER_HEADER, None):
            with reporter.step("Check others access to objects without attributes"):
                assert_access_to_container(
                    no_attributes_access[match_type],
                    other_wallet,
                    container,
                    objects_without_attributes.pop(),
                    file_path,
                    self.shell,
                    self.cluster,
                    xhdr=xhdr,
                )

            with reporter.step("Check others have full access to objects without deny attribute"):
                assert_access_to_container(
                    allowed_access[match_type], other_wallet, container, allow_objects.pop(), file_path, self.shell, self.cluster, xhdr=xhdr
                )

            with reporter.step("Check others have no access to objects with deny attribute"):
                with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                    head_object(other_wallet, container, deny_objects[0], self.shell, rpc_endpoint, xhdr=xhdr)

                with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                    get_object_from_random_node(other_wallet, container, deny_objects[0], self.shell, self.cluster, xhdr=xhdr)

            with reporter.step("Check others have access to objects with deny attribute and using bearer token"):
                assert_full_access_to_container(
                    other_wallet, container, deny_objects.pop(), file_path, self.shell, self.cluster, bearer, xhdr
                )

        allow_attribute = self.OTHER_HEADER if match_type == ape.MatchType.EQUAL else self.HEADER
        with reporter.step("Check others can PUT objects without denied attribute"):
            put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, attributes=allow_attribute)

        deny_attribute = self.HEADER if match_type == ape.MatchType.EQUAL else self.OTHER_HEADER
        with reporter.step("Check others can not PUT objects with denied attribute"):
            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, attributes=deny_attribute)

        with reporter.step("Check others can PUT objects with denied attribute and using bearer token"):
            put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, bearer_put, attributes=deny_attribute)

    @allure.title("Operations with allow APE rule with resource filters (match_type={match_type}, obj_size={object_size})")
    @pytest.mark.parametrize("match_type", [ape.MatchType.EQUAL, ape.MatchType.NOT_EQUAL])
    @pytest.mark.parametrize("object_size", ["simple"], indirect=True)
    @pytest.mark.container(ContainerSpec(basic_acl="0", allow_owner_via_ape=True))
    def test_ape_allow_filters_object(
        self,
        frostfs_cli: FrostfsCli,
        other_wallet: WalletInfo,
        container: str,
        objects_with_attributes: list[str],
        objects_with_other_attributes: list[str],
        objects_without_attributes: list[str],
        match_type: ape.MatchType,
        rpc_endpoint: str,
        file_path: TestFile,
        temp_directory: str,
    ):
        if match_type == ape.MatchType.EQUAL:
            allow_objects = objects_with_attributes
            deny_objects = objects_with_other_attributes
            allow_attribute = self.HEADER
            deny_attribute = self.OTHER_HEADER
            no_attributes_match_context = pytest.raises(Exception, match=OBJECT_ACCESS_DENIED)
        else:
            allow_objects = objects_with_other_attributes
            deny_objects = objects_with_attributes
            allow_attribute = self.OTHER_HEADER
            deny_attribute = self.HEADER
            no_attributes_match_context = expect_not_raises()

        with reporter.step("Allow operations for others except few operations by resource condition via APE"):
            resource_condition = ape.Condition('"check_key"', '"check_value"', ape.ConditionType.RESOURCE, match_type)
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            deny_rule = ape.Rule(ape.Verb.ALLOW, self.RESOURCE_OPERATIONS, [resource_condition, role_condition])

            frostfs_cli.ape_manager.add(
                rpc_endpoint, deny_rule.chain_id, target_name=container, target_type="container", rule=deny_rule.as_string()
            )

        with reporter.step("Wait for one block"):
            self.wait_for_blocks()

        with reporter.step("Check GET, PUT and HEAD operations with objects without attributes for OTHERS role"):
            oid = objects_without_attributes.pop()
            with no_attributes_match_context:
                assert head_object(other_wallet, container, oid, self.shell, rpc_endpoint)

            with no_attributes_match_context:
                assert get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster)

            with no_attributes_match_context:
                assert put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster)

        with reporter.step("Create bearer token with everything allowed for others role"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            rule = ape.Rule(ape.Verb.ALLOW, ALL_OBJECT_OPERATIONS, role_condition)
            bearer = create_bearer_token(frostfs_cli, temp_directory, container, rule, rpc_endpoint)

        with reporter.step("Check others can get and put objects without attributes and using bearer token"):
            oid = objects_without_attributes[0]
            with expect_not_raises():
                head_object(other_wallet, container, oid, self.shell, rpc_endpoint, bearer)

            with expect_not_raises():
                get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster, bearer)

            with expect_not_raises():
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, bearer)

        with reporter.step("Check others can get and put objects with attributes matching the filter"):
            oid = allow_objects.pop()
            with expect_not_raises():
                head_object(other_wallet, container, oid, self.shell, rpc_endpoint)

            with expect_not_raises():
                get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster)

            with expect_not_raises():
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, attributes=allow_attribute)

        with reporter.step("Check others cannot get and put objects without attributes matching the filter"):
            oid = deny_objects[0]
            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                head_object(other_wallet, container, oid, self.shell, rpc_endpoint)

            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                assert get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster)

            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                assert put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, attributes=deny_attribute)

        with reporter.step("Check others can get and put objects without attributes matching the filter with bearer token"):
            oid = deny_objects.pop()
            with expect_not_raises():
                head_object(other_wallet, container, oid, self.shell, rpc_endpoint, bearer)

            with expect_not_raises():
                get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster, bearer)

            with expect_not_raises():
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, bearer, attributes=allow_attribute)

    @allure.title("PUT and GET object using bearer with objectID in filter (obj_size={object_size}, match_type=NOT_EQUAL)")
    @pytest.mark.container(ContainerSpec(basic_acl="0", allow_owner_via_ape=True))
    def test_ape_filter_object_id_not_equals(
        self,
        frostfs_cli: FrostfsCli,
        default_wallet: WalletInfo,
        other_wallet: WalletInfo,
        container: str,
        temp_directory: str,
        file_path: TestFile,
    ):
        with reporter.step("Put object to container"):
            oid = put_object_to_random_node(default_wallet, file_path, container, self.shell, self.cluster)

        with reporter.step("Create bearer token with objectID filter"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            object_condition = ape.Condition.by_object_id(oid, ape.ConditionType.RESOURCE, ape.MatchType.NOT_EQUAL)
            rule = ape.Rule(ape.Verb.ALLOW, ALL_OBJECT_OPERATIONS, [role_condition, object_condition])
            bearer = create_bearer_token(frostfs_cli, temp_directory, container, rule, self.cluster.default_rpc_endpoint)

        with reporter.step("Others should be able to put object using bearer token"):
            with expect_not_raises():
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, bearer)

        with reporter.step("Others should not be able to get object matching the filter"):
            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster, bearer)

    @allure.title("PUT and GET object using bearer with objectID in filter (obj_size={object_size}, match_type=EQUAL)")
    @pytest.mark.container(ContainerSpec(basic_acl="0", allow_owner_via_ape=True))
    def test_ape_filter_object_id_equals(
        self,
        frostfs_cli: FrostfsCli,
        default_wallet: WalletInfo,
        other_wallet: WalletInfo,
        container: str,
        temp_directory: str,
        file_path: TestFile,
    ):
        with reporter.step("Put object to container"):
            oid = put_object_to_random_node(default_wallet, file_path, container, self.shell, self.cluster)

        with reporter.step("Create bearer token with objectID filter"):
            role_condition = ape.Condition.by_role(ape.Role.OTHERS)
            object_condition = ape.Condition.by_object_id(oid, ape.ConditionType.RESOURCE, ape.MatchType.EQUAL)
            rule = ape.Rule(ape.Verb.ALLOW, ALL_OBJECT_OPERATIONS, [role_condition, object_condition])
            bearer = create_bearer_token(frostfs_cli, temp_directory, container, rule, self.cluster.default_rpc_endpoint)

        with reporter.step("Others should not be able to put object using bearer token"):
            with pytest.raises(Exception, match=OBJECT_ACCESS_DENIED):
                put_object_to_random_node(other_wallet, file_path, container, self.shell, self.cluster, bearer)

        with reporter.step("Others should be able to get object matching the filter"):
            with expect_not_raises():
                get_object_from_random_node(other_wallet, container, oid, self.shell, self.cluster, bearer)
