import logging

import allure
import pytest
from frostfs_testlib import reporter
from frostfs_testlib.resources.error_patterns import OBJECT_NOT_FOUND
from frostfs_testlib.steps.cli.object import get_object_from_random_node, head_object, put_object_to_random_node
from frostfs_testlib.steps.epoch import get_epoch
from frostfs_testlib.storage.dataclasses.object_size import ObjectSize
from frostfs_testlib.storage.dataclasses.wallet import WalletInfo
from frostfs_testlib.testing.cluster_test_base import ClusterTestBase
from frostfs_testlib.testing.test_control import expect_not_raises
from frostfs_testlib.utils.file_utils import TestFile

from ...helpers.container_spec import ContainerSpecs
from ...helpers.utility import wait_for_gc_pass_on_storage_nodes

logger = logging.getLogger("NeoLogger")


@pytest.mark.nightly
@pytest.mark.sanity
@pytest.mark.grpc_api
class TestObjectApiLifetime(ClusterTestBase):
    @pytest.mark.container(ContainerSpecs.PublicReadWrite)
    @allure.title("Object is removed when lifetime expired (obj_size={object_size})")
    def test_object_api_lifetime(self, container: str, test_file: TestFile, default_wallet: WalletInfo, object_size: ObjectSize):
        """
        Test object deleted after expiration epoch.
        """

        wallet = default_wallet

        epoch = get_epoch(self.shell, self.cluster)

        oid = put_object_to_random_node(wallet, test_file.path, container, self.shell, self.cluster, expire_at=epoch + 1)
        with expect_not_raises():
            head_object(wallet, container, oid, self.shell, self.cluster.default_rpc_endpoint)

        with reporter.step("Tick two epochs"):
            self.tick_epochs(2)

        # Wait for GC, because object with expiration is counted as alive until GC removes it
        wait_for_gc_pass_on_storage_nodes()

        with reporter.step("Check object deleted because it expires on epoch"):
            with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                head_object(wallet, container, oid, self.shell, self.cluster.default_rpc_endpoint)
            with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                get_object_from_random_node(wallet, container, oid, self.shell, self.cluster)

        with reporter.step("Tick additional epoch"):
            self.tick_epoch()

        wait_for_gc_pass_on_storage_nodes()

        with reporter.step("Check object deleted because it expires on previous epoch"):
            with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                head_object(wallet, container, oid, self.shell, self.cluster.default_rpc_endpoint)
            with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                get_object_from_random_node(wallet, container, oid, self.shell, self.cluster)
