import allure
import pytest
from frostfs_testlib import reporter
from frostfs_testlib.shell import Shell
from frostfs_testlib.steps.cli.container import create_container, delete_container, get_container, list_containers
from frostfs_testlib.steps.session_token import ContainerVerb, get_container_signed_token
from frostfs_testlib.storage.dataclasses.wallet import WalletInfo
from frostfs_testlib.testing.cluster_test_base import ClusterTestBase


@pytest.mark.nightly
@pytest.mark.static_session_container
class TestSessionTokenContainer(ClusterTestBase):
    @pytest.fixture(scope="module")
    def static_sessions(
        self,
        owner_wallet: WalletInfo,
        user_wallet: WalletInfo,
        client_shell: Shell,
        temp_directory: str,
    ) -> dict[ContainerVerb, str]:
        """
        Returns dict with static session token file paths for all verbs with default lifetime
        """
        return {verb: get_container_signed_token(owner_wallet, user_wallet, verb, client_shell, temp_directory) for verb in ContainerVerb}

    @allure.title("Static session with create operation")
    def test_static_session_token_container_create(
        self,
        owner_wallet: WalletInfo,
        user_wallet: WalletInfo,
        static_sessions: dict[ContainerVerb, str],
    ):
        """
        Validate static session with create operation
        """
        with reporter.step("Create container with static session token"):
            cid = create_container(
                user_wallet,
                session_token=static_sessions[ContainerVerb.CREATE],
                shell=self.shell,
                endpoint=self.cluster.default_rpc_endpoint,
                wait_for_creation=False,
            )

        container_info: dict[str, str] = get_container(owner_wallet, cid, shell=self.shell, endpoint=self.cluster.default_rpc_endpoint)
        assert container_info["ownerID"] == owner_wallet.get_address()

        assert cid not in list_containers(user_wallet, shell=self.shell, endpoint=self.cluster.default_rpc_endpoint)
        assert cid in list_containers(owner_wallet, shell=self.shell, endpoint=self.cluster.default_rpc_endpoint)

    @allure.title("[NEGATIVE] Static session without create operation")
    def test_static_session_token_container_create_with_other_verb(
        self,
        user_wallet: WalletInfo,
        static_sessions: dict[ContainerVerb, str],
    ):
        """
        Validate static session without create operation
        """
        with reporter.step("Try create container with static session token without PUT rule"):
            for verb in [verb for verb in ContainerVerb if verb != ContainerVerb.CREATE]:
                with pytest.raises(Exception):
                    create_container(
                        user_wallet,
                        session_token=static_sessions[verb],
                        shell=self.shell,
                        endpoint=self.cluster.default_rpc_endpoint,
                        wait_for_creation=False,
                    )

    @allure.title("[NEGATIVE] Static session with create operation for other wallet")
    def test_static_session_token_container_create_with_other_wallet(
        self,
        stranger_wallet: WalletInfo,
        static_sessions: dict[ContainerVerb, str],
    ):
        """
        Validate static session with create operation for other wallet
        """
        with reporter.step("Try create container with static session token without PUT rule"):
            with pytest.raises(Exception):
                create_container(
                    stranger_wallet,
                    session_token=static_sessions[ContainerVerb.CREATE],
                    shell=self.shell,
                    endpoint=self.cluster.default_rpc_endpoint,
                    wait_for_creation=False,
                )

    @allure.title("Static session with delete operation")
    def test_static_session_token_container_delete(
        self,
        owner_wallet: WalletInfo,
        user_wallet: WalletInfo,
        static_sessions: dict[ContainerVerb, str],
    ):
        """
        Validate static session with delete operation
        """
        with reporter.step("Create container"):
            cid = create_container(
                owner_wallet,
                shell=self.shell,
                endpoint=self.cluster.default_rpc_endpoint,
                wait_for_creation=False,
            )
        with reporter.step("Delete container with static session token"):
            delete_container(
                wallet=user_wallet,
                cid=cid,
                session_token=static_sessions[ContainerVerb.DELETE],
                shell=self.shell,
                endpoint=self.cluster.default_rpc_endpoint,
                await_mode=True,
            )

        assert cid not in list_containers(owner_wallet, shell=self.shell, endpoint=self.cluster.default_rpc_endpoint)
